@@ -1,7 +1,7 @@
 """Top-level package for IceNet."""
 
-__author__ = "Tom Andersson / James Byrne"
+__author__ = "British Antarctic Survey / Alan Turing Institute"
 __copyright__ = "British Antarctic Survey"
 __email__ = "jambyr@bas.ac.uk"
 __license__ = "MIT"
-__version__ = "0.2.0"
+__version__ = "0.2.2"
